/*
 * Croquis basé entre autres sur l'exemple repris dans la librairie accompagnant le display I2C/TWI LCD1602 de DFROBOT, article TOY0046.

 * Version V0.60c4L 03/02/2017 : ajout de l'enregistrement des adresse des locos dans l'EEPROM

GENERALITES
 * Ce croquis pour Arduino UNO décrit un contrôleur destiné à commander la centrale DCC++ telle que proposée par Gregg E. Berman version 1.2.1+.
 * Elle envoie via la ligne série (RX/TX) les messages standardisés à destination de la centrale pour la commande des locos;
 * par exemple, pour une commande de traction (loco registre 1, adresse decodeur 62, vitesse cran 50, sens marche 1 avant), le message envoyé sera <t 1 62 50 1>; les signes < et > font partie intégrante du message.
 * Le signal à destination de la centrale DCC++ part de la pin 1 TX; elle est a raccorder à la pin 2 RX de la carte de la centrale DCC++
 * Attention! Retirer la liaison TX/RX pendant les téléversements sinon on obtient des erreurs en fin de compilation
 * Le contrôleur est destiné à commander 4 locos; chaque loco est appelée par son propre bouton; les adresses DCC des locos sont enregistrées en EEPROM.
 * Pour chaque loco, les fonctions F0 éclairage et F1 a F4(ON/OFF) sont également commandées par leur propre bouton (boutons de couleur).
 * Ce contrôleur permet également de configurer sur voie principale les CV des décodeurs. Il ne lit pas les valeurs des CV.
 * La programmation ayant lieu sur voie principale ne peut donc se réaliser qu'avec une seule loco présente sur la voie.
 * La commande de vitesse des locos est originale car elle peut se faire selon 2 manières différentes (traditionnelle ou rapide ; voir "commande des locos" ci-dessous).

TELECOMMANDE INFRAROUGE (IR)
 * Avec télécommande infrarouge PANASONIC DVD/TV
 * Les codes pour une télécommande de TV SONY RM-ED050 sont également fournis; ceux-ci pourraient également fonctionner avec d'autres télécommandes similaires de la même marque.
 * Choisissez l'un des deux sets en commentant ou décommentant un jeu de codes (un seul doit rester actif); vous pouvez également consituer votre propre jeu de codes.
 * Il n'y a évidemment qu'un seul jeu de code qui doit rester actif!

 * Attention! Certaines télécommandes envoient un (des) code(s) spécifique(s) lors de l'appui long ou du maintien sur une touche;
 * dans ce croquis ces codes sont identifiés par repeat1 et repeat2; ces deux codes sont utilisés pour envoyer des répétitions de commande
 * lors de l'appui sur les flèches haut ou bas (pour permettre le réglage fin de la vitesse 'en rafale').

 * Le capteur IR TSOP 4838 est raccordé comme suit à l'Arduino UNO :
 * - Pin 1 (à gauche lorsque le capteur est vu de face) = Vout (pin 3 de l'Arduino)
 * - Pin 2 (au centre) = GND raccordée à un GND de l'Arduino
 * - Pin 3 (à droite lorsque le capteur est vu de face) = Vcc (+5v de l'Arduino)

AFFICHAGE
 * L'affichage se réalise à l'aide d'un LCD 2 lignes 16 caractères I2C LCD1602 de DFROBOT, article TOY0046.
 * Le display I2C LCD1602 est raccordé comme suit : SDA à la pin analog4 et SCL à la pin analog5 de l'Arduino UNO; attention dans certaines documentations, ces 2 pins sont inversées à tort.

REPETITION SONORE DES COMMANDES
 * Pour la facilité d'utilisation, la réception de certaines commandes sont confirmées par un bip sonore :
 * - commande du sens de marche : un bip aigu
 * - sélection d'une loco : bip grave; le nombre de bips correspond au numéro d'ordre de la loco sélectionnée (1 bip pour loco 1,  ..., 4 bips pour loco 4)
 * - confirmation de l'enregistrement de CV (1 bip aigu) ou de leur valeur (2 bips aigus) et de l'envoi de la commande à la centrale (3 bips graves).
 * - association d'une adresse décodeur à un bouton loco.
 * Le buzzer (ou HP) est connecté entre GND et pin 8.
 * Pour éviter des conflits de timer avec la reception IR, on utilise la librairie NewTone

MENUS
 * Il y a 4 menus disponibles :
 * - menu 0 : à la mise sous tension du système -> écran d'accueil et numéro de version ; pas de signal DCC envoyé.
 * - menu 1 : menu RUN (appui sur la touche PLAY) -> menu normal de circulation des locos ; au début, la loco 1 est sélectionnée par défaut
 * - menu 2 : menu PROGRAM (appui sur la touche PROG) -> menu de programmation des CV du décodeur; une seule loco sur la voie! Annulation par la touche DELETE
 * - menu 3 : menu ADRESSE (appui sur la touche INPUT SELECT) -> menu de modification des adresses DCC des locos activées par les boutons 1 à 4.

COMMANDE DES LOCOS
 * Pour commander une loco :
 * - sélectionnez le menu 1 en appuyant sur PLAY (à faire une fois en début d'exploitation)
 * - sélectionnez la touche correspondante à la loco (touches loco1 à loco4)que vous voulez commander
 * - sélectionnez le sens de marche en appuyant sur les "flèche gauche" ou "flèche droite"
 * - sélectionnez la vitesse souhaitée;
 *   il y a deux moyens de régler la vitesse :
 *   + commande TRADITIONNELLE : appuyez sur la "flèche haut" pour augmenter la vitesse et "flèche bas" pour la diminuer; la variation a lieu cran par cran.
 *   + commande RAPIDE : appuyez sur un chiffre de 0 à 9 -> appui sur 1=cran 15, 2=cran 25, 3=cran35 ,..., 9=cran 95 ; 0=arrêt;
 *     cette commande rapide permet d'obtenir immédiatement une vitesse prédéfinie en épargnant les appuis sur les touches "flèche haut" et "flèche bas".
 *     Les deux commandes peuvent être utilisées simultanément et se complètent naturellement; par exemple, une loco entre en gare (appui sur la touche 4 => cran 45 pour passer sur l'aiguillage en position déviée),
 *     puis approche de l'extrémité du quai (appui sur la touche 1 => cran 15) et l'arrêt a lieu par appuis successifs sur la "flèche bas" (diminution cran par cran) ou appui sur la touche 0.
 *     L'appui sur la touche 0 garantit une vitesse nulle.
 * - pour allumer ou éteindre les feux, appuyez sur la touche "Lights" (touche jaune sur la télécommande PANASONIC); c'est une fonction bascule ON/OFF ;
 * - faites de même pour les autres fonctions ; en standard, les fonctions F1 à F3 sont implémentées et activées par un bouton de couleur F1=rouge F2=vert F3=bleu)

 MODIFICATION DES CV DU DECODEUR D'UNE LOCO
 * Attention : une seule loco sur le circuit pendant la programmation!
 * Pour modifier une CV :
 * - sélectionnez le menu 2 en appuyant sur PROGRAM;
 * - sélectionnez le numéro de la CV a modifier par le clavier numerique; ne pas introduire les zeros non significatifs; confirmez par OK
 * - sélectionnez la valeur à affecter a cette CV et confirmez par OK;
 * - après chaque appui sur OK, il y a confirmation sonore de la sélection.

 MODIFICATION DES ADRESSES DCC
 * Pour modifier l'adresse d'une loco activée par un bouton déterminé :
 * - sélectionnez le menu 3 en appuyant sur INPUT SELECT ;
 * - en appuyant sur les flèches gauche ou droite, vous pouvez consulter les adresses affectées à chaque bouton (LOC);
 * - sélectionnez le numéro du bouton souhaité en appuyant sur la touche OK ;
 * - le numéro du bouton (LOC), l'adresse actuelle (NOW) et la nouvelle adresse (CHANGE? ->provisoirement à 0) sont affichés ;
 * - introduisez la nouvelle adresse souhaitée pour ce bouton et confirmez par OK;
 * - le nouvel état s'affiche, accompagné d'une confirmation sonore.

DETECTION DES ERREURS ET AFFICHAGE DES CODES INFRAROUGES
 * Pour afficher les valeurs numériques des codes reçus par la carte sur le moniteur série, décommentez la ligne //#define affichage_code;
 * ceci permet de contrôler le fonctionnement de votre télécommande ainsi que la correspondance entre les touches et les codes prédéfinis.
 * Pour annuler cette fonctionnalité, commentez la ligne "#define affichage_code";
 * Pour contrôler le déroulement du programme, décommentez la ligne "//#define mode_debug" avant de téléverser le croquis; certains messages s'afficheront sur le moniteur série.
 * En fonctionnement normal, #define affichage_code et #define mode_debug doivent être commentées sinon le controleur fonctionnera de manière anormale.

* Copyright (c) 2017 Daniel PONTHIERE & Locoduino
* This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation, either version 3 of the License, or at your option) any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU General Public License for more details. <http://www.gnu.org/licenses/>.
*/

/*******************CODE********************/

#include <IRLib.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <NewTone.h>
#include <EEPROM.h>

#define ver "v0.60c4L by DPo"

//#define mode_debug //compilation en mode debug; pour annuler placer // devant le #
//#define affichage_code //compilation en mode affichage de code : affiche les codes IR sur le moniteur serie; pour annuler placer // devant le #

//definition des codes de la telecommande Panasonic

//codes pour télécommande Panasonic librairie IRLib
#define arrow_up 3796283338
#define repeat1 2166136261 // certaines télécommandes envoient un (des) code(s) spécifique(s) lors de l'appui long ou du maintien sur une touche
#define repeat2 1556999678// certaines télécommandes envoient un (des) code(s) spécifique(s) lors de l'appui long ou du maintien sur une touche
#define arrow_down 989538004
#define arrow_left 1296259434
#define arrow_right 3138313506
#define digit_0 950179491
#define digit_1 1366498449
#define digit_2 1242645753
#define digit_3 577270721
#define digit_4 278230305
#define digit_5 3236444089
#define digit_6 2050858721
#define digit_7 1376882089
#define digit_8 1077841673
#define digit_9 809146499
#define button_pause 2860196199
#define button_play 3737721253
#define button_light 1562762607 // jaune fonction F0
#define button_OK 189926636
#define button_PROG 618970697
#define button_DELETE 52104363
#define button_loco1 3096697063
#define button_loco2 4003586079
#define button_loco3 1913490470
#define button_loco4 3857854961
#define button_F1 568341793 //rouge
#define button_F2 938060567 //vert
#define button_F3 645776457 //bleu
#define button_F4 309669706
#define button_F5 400358607
#define button_F6 191349048
#define button_adresseLoco 3527880061//INPUT SELECT

/* codes pour télécommande TV SONY RM-ED050
#define arrow_up 3962033213
#define repeat1 425423268 // certaines télécommandes envoient un (des) code(s) spécifique(s) lors de l'appui long ou du maintien sur une touche
#define repeat2 1556999678// certaines télécommandes envoient un (des) code(s) spécifique(s) lors de l'appui long ou du maintien sur une touche ; pas utilisee ici
#define arrow_down 141285788
#define arrow_left 440544835
#define arrow_right 2721830948
#define digit_0 990610224
#define digit_1 544993039
#define digit_2 3896860046
#define digit_3 3417827657
#define digit_4 740241778
#define digit_5 2422134005
#define digit_6 2363614204
#define digit_7 828053765
#define digit_8 3330015670
#define digit_9 3021527401
#define button_pause 920158043
#define button_play 1664997572
#define button_light 3234147071 //bouton jaune fonction F0
#define button_OK 1922410756  //bouton central
#define button_PROG 726507874 //bouton record point rouge
#define button_DELETE 3934206650  //bouton EXIT
#define button_loco1 583823042 //bouton retour debut |<<
#define button_loco2 735876127 //bouton retour <<
#define button_loco3 3173596934 //bouton avance >>
#define button_loco4 1492441743 //bouton avance fin >>|
#define button_F1 3795534161 // bouton rouge
#define button_F2 3461693924 //bouton vert
#define button_F3 2046533204 //bouton bleu
#define button_F4 191349048
#define button_F5 400358607
#define button_F6 191349048
#define button_adresseLoco 91349048//bouton DIGITAL
*/

LiquidCrystal_I2C lcd(0x20, 16, 2); // attention! Lorsque tous les 3 jumpers sont placés (réglage d'usine) l'addresse LCD est 0x20
byte IRpin = 3;

IRrecv My_Receiver(IRpin);
IRdecode My_Decoder;
IRdecodeHash My_Hash_Decoder;

//déclaration tableau pour 4 loco actives

const int maxLocos = 4;// nombre max de locos
const int maxFonctions = 5;//nombre max de fonctions
int LocoAdresse [maxLocos];
int LocoDirection [maxLocos] = {1, 1, 1, 1};
int LocoSpeed [maxLocos] = {0, 0, 0, 0};
byte LocoFonction [maxLocos][maxFonctions];//tableau locos/fonctions
int CommandeFonction[maxLocos];//tableau des commandes de fonction a envoyer pour chaque loco
int AdresseActive = 0; // numero de la loco active (0 par defaut)

//définition des paramètres sonores
byte tone_pin = 8; //pin a laquelle est raccordé le HP ou buzzer
int frequence_A1 = 1000; //bip a frequence aigue 1 (Hz)
int duree_courte = 10;//duree courte (msec)
int frequence_G1 = 100; //bip a frequence grave 1 (Hz)
int duree_moyenne = 50;// duree moyenne (msec)

byte menu = 0; //menu=0 => arret ;menu = 1 =>RUN;menu = 2 =>PROGRAM; menu = 3 =>ADRESSE (lien adresses/boutons des locos)
byte bouton_valeur = 0;
int digit_C = 0;
byte digit_D = 0;
byte digit_U = 0;
int digit_COMMAND = 0;
int comptageAdresse = 0; //index de comptage pour menu 3
byte changement_adresse = 0; //passage en mode changement adresse dans menu 3

int timerefresh = 2000 ;//frequence de rafraichissement des fonctions (msec).

int vit_abs = 0; //valeur absolue de la vitesse
int sens_marche = 1; // seulement deux valeurs possibles de sens de marche : 1= AV, 0 = AR

int long time = 0; // mesure du temps avant répétition
int timebcl = 0; // mesure du temps de boucle
int oldtimebcl = 0; // mesure du temps de boucle

int long time_light = 0;//controle du temps entre deux commandes des feux (antirebond)
int long time_button = 0;//controle du temps entre deux commandes ir (antirebond)

byte CV_NUM = 0;
byte CV_VAL = 0;

int CV_Valeur = 0;
int CV_Numero = 0;

byte adresse_NUM = 0;
int adresse_Numero = 0;


void setup()
{
  pinMode(tone_pin, OUTPUT);//definition de la pin de production des bips sonores
  lcd.init();              // initialisation du LCD
  lcd.backlight();	   // allumage retroeclairage

  Serial.begin(115200);

  My_Receiver.enableIRIn();// activation de la reception IR
  //lecture des adresses des locos
  lectureAdresse();
  affichageMenu(menu); //ecran affichage en fonction du menu
}// fin de SETUP

void loop()
{
  // Controle de reception d'un code IR
  if (My_Receiver.GetResults(&My_Decoder) && millis() > time_button + 250)
  {
    My_Hash_Decoder.copyBuf(&My_Decoder);//copie des resultats dans le decodeur
    My_Decoder.decode();

    lcd.setCursor(0, 1);

    My_Hash_Decoder.decode();

#ifdef affichage_code
    Serial.println(" Code reçu => ");
    Serial.println(My_Hash_Decoder.hash, DEC); // Impression du code ir recu
#endif

    switch (My_Hash_Decoder.hash)
    {
      case arrow_up:
        bouton_valeur = 12;
        break;

      case arrow_down:
        bouton_valeur = 13;
        break;

      case arrow_left:
        bouton_valeur = 10;
        break;

      case arrow_right:
        bouton_valeur = 11;
        break;

      case digit_0:
        bouton_valeur = 0;
        break;

      case digit_1:
        bouton_valeur = 1;
        break;

      case digit_2:
        bouton_valeur = 2;
        break;

      case digit_3:
        bouton_valeur = 3;
        break;

      case digit_4:
        bouton_valeur = 4;
        break;

      case digit_5:
        bouton_valeur = 5;
        break;

      case digit_6:
        bouton_valeur = 6;
        break;

      case digit_7:
        bouton_valeur = 7;
        break;

      case digit_8:
        bouton_valeur = 8;
        break;

      case digit_9:
        bouton_valeur = 9;
        break;

      case button_pause:// meme fonction que button_stop
        bouton_valeur = 24;
        break;

      case button_play:
        bouton_valeur = 25;
        break;

      case button_light:
        if (millis() > time_light + 200)// antirebond : si moins de 200 msec on ne tient pas compte de l'appui sur le bouton
        {
          time_light = millis();
        }
        bouton_valeur = 40;
        break;

      case button_OK:
        bouton_valeur = 20;
        break;

      case button_PROG:
        bouton_valeur = 21;
        break;

      case button_adresseLoco:
        bouton_valeur = 22;
        break;

      case button_loco1:
        bouton_valeur = 26;
        break;

      case button_loco2:
        bouton_valeur = 27;
        break;

      case button_loco3:
        bouton_valeur = 28;
        break;

      case button_loco4:
        bouton_valeur = 29;
        break;

      case button_DELETE:
        bouton_valeur = 30;
        break;

      case button_F1:
        bouton_valeur = 41;
        break;

      case button_F2:
        bouton_valeur = 42;
        break;

      case button_F3:
        bouton_valeur = 43;
        break;

      case button_F4:
        bouton_valeur = 44;
        break;

      case repeat1:
        if (bouton_valeur != 12 && bouton_valeur != 13)//pas de repetion de touche sauf pour fleche haut et fleche bas
        {
          bouton_valeur = 99;
        }
        break;

      case repeat2:
        if (bouton_valeur != 12 && bouton_valeur != 13)//pas de repetion de touche sauf pour fleche haut et fleche bas
        {
          bouton_valeur = 99;
        }
        break;

      default :
        bouton_valeur = 99;
        break;

    } //end switch

#ifdef affichage_code
    Serial.print ("Bouton_valeur  ->");//pour mise au point
    Serial.println(bouton_valeur);//idem
#endif

    //Determination des actions a entreprendre en fonction de la valeur du bouton activé
    if (bouton_valeur > 19) //changement de menu ou bouton special
    {
      switch (bouton_valeur)
      {

        case 24:
          lcd.clear();
          affichageMenu(0);
          menu = 0; //Arret DCC
          break;

        case 25:
          lcd.clear();
          affichageMenu(1);
          menu = 1; //RUN
          break;

        case 22:
          if (menu != 3)
          {
            int i = 0;
            lcd.clear();
            affichageMenu(3);
            menu = 3; //mapping entre les adresses loco et les boutons

          }//End if menu!=3
          break;

        case 26://selection loco 1
          changementLoco(0);
          break;

        case 27://selection loco 2
          changementLoco(1);
          break;

        case 28://selection loco 3
          changementLoco(2);
          break;

        case 29://selection loco 4
          changementLoco(3);
          break;

        case 40://selection F0
          Activation_Fonction(0);
          break;

        case 41://selection F1
          Activation_Fonction(1);
          break;

        case 42://selection F2
          Activation_Fonction(2);
          break;

        case 43://selection F3
          Activation_Fonction(3);
          break;

        case 44://selection F4
          Activation_Fonction(4);
          break;

        case 21:
          if (menu != 2)
          {
            lcd.clear();
            menu = 2; //PROGRAM
            CV_Valeur = 0;
            CV_Numero = 0;
            digit_C = 0;
            digit_D = 0;
            digit_U = 0;
            digit_COMMAND = 0;
            affichageMenu(2);
          }//END menu !=2
          break;

      }// END SWITCH

    }// END IF changement de menu/bouton special
    //menu RUN
    if (menu == 1) //exploitation normale
    {

      LocoVitesse();// fonction de determination de la vitesse de la loco
      if (vit_abs == 0)//arret normal => afficher vitesse 0
      {
        affichage_LCD_INT(8, 0, 0);
      }
      else
      {
        affichage_LCD_INT(8, 0, vit_abs);
      }

      affichage_sens_marche (sens_marche);
    }//ENF IF menu=1


    //menu PROGRAM de configuration des CV
    if (menu == 2)
    {
      if (bouton_valeur != 30)//a executer sauf si appui sur bouton DELETE
      {
#ifdef mode_debug
        Serial.print("Bouton?  ");
        Serial.print(bouton_valeur);
        Serial.print ("Valeur avant if  ");
        Serial.print (digit_C);
        Serial.print(" D : ");
        Serial.print(digit_D);
        Serial.print(" U : ");
        Serial.print(digit_U);
        Serial.print(" CV_NUM = ");
        Serial.println(CV_NUM);
#endif

        if (bouton_valeur != 20) //tant qu'on a pas appuye sur OK, on introduit les chiffres dans digit_COMMAND
        {
          digit_C = digit_D * 10;// on decale les valeurs d'une position vers la gauche a chaque tour
          digit_D = digit_U * 10;// idem
          if (bouton_valeur < 10)// a faire uniquement pour les valeurs chiffrees
          {
            digit_U = bouton_valeur;
          }
          digit_COMMAND = digit_C + digit_D + digit_U;//constitution du numero complet

          delay(200);// pour eviter que les valeurs soient prises en compte plus d'une fois
          lcd.setCursor (1, 0);
          lcd.print(digit_COMMAND);// affichage de la valeur actuellement enregistree dans digit_COMMAND
        }

        if (bouton_valeur == 20 && CV_NUM == 0 && digit_COMMAND > 0) // appui sur OK ou ENTER lors de la selection CV_NUM
        {
          CV_Numero = digit_COMMAND;// on enregistre digit_COMMAND comme numero de CV
          CV_NUM = 1;// memorise le fait qu'on a enregistre un numero de CV
          lcd.setCursor (13, 0);// affiche le numero de CV
          lcd.print(digit_COMMAND);
          bip(tone_pin, frequence_A1, duree_courte, 1); // confirmation sonore introduction numero
          digit_COMMAND = 0;// reset des valeurs
          digit_C = 0;
          digit_D = 0;
          digit_U = 0;

          bouton_valeur = 21;// on force la valeur a bouton program pour rester dans la boucle de programmation
        }// END IF appui sur OK lors de la selection CV_NUM

        if (bouton_valeur == 20 && CV_VAL == 0 && CV_NUM == 1) // appui sur OK ou ENTER lors de la selection CV_VAL (numero de CV deja enregistre)
        {
          CV_Valeur = digit_COMMAND;//on enregistre digit_COMMAND comme valeur de CV
          CV_NUM = 0;
          CV_VAL = 0;

          lcd.setCursor (13, 0);
          lcd.print(CV_Numero);
          lcd.setCursor (13, 1);
          lcd.print(digit_COMMAND);
          bip(tone_pin, frequence_A1, duree_courte, 2); // confirmation sonore de l'introduction d'un numero (2 bips)
          digit_COMMAND = 0;

          // envoi de la commande serie DCC++ vers la centrale

          Serial.print("<1>");
          Serial.print("<w ");
          Serial.print(LocoAdresse[AdresseActive]);//numéro de loco
          Serial.print(" ");
          Serial.print(CV_Numero);
          Serial.print(" ");
          Serial.print(CV_Valeur);
          Serial.write(">");

          lcd.setCursor(4, 1);
          lcd.print("!");// confirmation envoi de la commande à la centrale
          bip(tone_pin, frequence_G1, duree_moyenne, 3); //confirmation sonore fin de processus
          delay(1000);
          lcd.print(" ");// reset de confirmation

          //reset des valeurs
          CV_Valeur = 0;
          CV_Numero = 0;
          digit_C = 0;
          digit_D = 0;
          digit_U = 0;

        }// ENDIF appui sur OK lors de la selection CV_NUM (numero CV deja enregistre)
      }
      else //appui sur bouton DELETE
      {
        //reset des valeurs
        CV_Valeur = 0;
        CV_Numero = 0;
        digit_C = 0;
        digit_D = 0;
        digit_U = 0;
        menu = 1; //retour en mode RUN
        vit_abs = 0;
        affichageMenu(menu);

      }//END IF bouton_valeur != 30 (tant qu'on appuie pas sur DELETE)
    }// END IF menu = 2
#ifdef mode_debug
    Serial.print (" numero menu : ");
    Serial.println (menu);
#endif

    //menu adresseLoco de correspondance entre les adresses et les boutons
    if (menu == 3)
    {
#ifdef mode_debug
      Serial.print(" Bouton Valeur ?  ");
      Serial.print(bouton_valeur);
      Serial.print(" debut de menu 3  ");
      Serial.print(" adresse_NUM=> ");
      Serial.print (adresse_NUM);
      Serial.print (" digit_COMMAND = ");
      Serial.println (digit_COMMAND);
#endif

      if (comptageAdresse < maxLocos)
      {
        affichageMenu(3);
        affichage_LCD_INT(3, 1, comptageAdresse + 1);
        affichage_LCD_INT(7, 1, LocoAdresse[comptageAdresse]);
        affichage_LCD_INT(14, 1, digit_COMMAND);
        if (bouton_valeur == 11) //fleche droite
        {
          comptageAdresse = comptageAdresse + 1;
          if (comptageAdresse == maxLocos)
          {
            comptageAdresse = maxLocos - 1; //limitation à maxLoco-1;
          }//fin comptageAdresse=maxLocos
        }//fin bouton_valeur==11

        if (bouton_valeur == 10) //fleche gauche
        {
          comptageAdresse = comptageAdresse - 1;
          if (comptageAdresse <= 0)
          {
            comptageAdresse = 0 ;//limitation à 0
          }//fin comptageAdresse<=0
        }//fin bouton_valeur==10
        //affichageMenu(3);
#ifdef mode_debug
        Serial.print(" Comptage adresse = ");
        Serial.println(comptageAdresse);
#endif
        if (bouton_valeur == 20 | changement_adresse == 1) //appui sur OK => début de changement d'adresse ou changement_adresse==1 => chgt en cours
        {
#ifdef mode_debug
          Serial.print(F("changement adresse"));
#endif
          changement_adresse = 1; //memorisation du passage en mode 'changement d'adresse'
          lcd.setCursor (10, 1);
          lcd.print(">");//indication du debut d'introduction de la nouvelle adresse
          // if (bouton_valeur != 30)//a executer sauf si appui sur bouton DELETE
          // {
#ifdef mode_debug
          Serial.print("ComptageAdresse = ");
          Serial.print(comptageAdresse);
          Serial.print(" Bouton?  ");
          Serial.print(bouton_valeur);
#endif

          if (bouton_valeur != 20 && bouton_valeur != 30) //tant qu'on a pas appuye sur OK, on introduit les chiffres dans digit_COMMAND
          {
#ifdef mode_debug
            Serial.print ("Introduction chiffres");
#endif
            digit_C = digit_D * 10;// on decale les valeurs d'une position vers la gauche a chaque tour
            digit_D = digit_U * 10;// idem
            if (bouton_valeur < 10)// a faire uniquement pour les valeurs chiffrees
            {
              digit_U = bouton_valeur;
            }
            digit_COMMAND = digit_C + digit_D + digit_U;//constitution du numero complet
            delay(200);// pour eviter que les valeurs soient prises en compte plus d'une fois
            affichage_LCD_INT(14, 1, digit_COMMAND);// affichage de la valeur actuellement enregistree dans digit_COMMAND
          }

          if (bouton_valeur == 20 && digit_COMMAND > 0 && bouton_valeur != 30) // appui sur OK ou ENTER lors de la selection d'adresse : confirmation
          {
#ifdef mode_debug
            Serial.println (F("confirmation"));
#endif
            lcd.setCursor (10, 1);
            lcd.print(" ");
            lcd.setCursor(15, 0);
            lcd.print("D");//effacement de l'indication du debut d'introduction de la nouvelle adresse
            adresse_Numero = digit_COMMAND;// on enregistre digit_COMMAND comme numero d'adresse
            LocoAdresse[comptageAdresse] = adresse_Numero;
            ecritureAdresse(comptageAdresse);
            affichage_LCD_INT(14, 1, adresse_Numero);
            bip(tone_pin, frequence_A1, duree_courte, 1); // confirmation sonore introduction numero
            digit_COMMAND = 0;// reset des valeurs
            digit_C = 0;
            digit_D = 0;
            digit_U = 0;
            changement_adresse = 0; // reset car changement realise
          }// END IF appui sur OK lors de la selection adresse_NUM


          if (bouton_valeur == 30) //appui sur bouton DELETE
          {
#ifdef mode_debug
            Serial.println ("delete!!!!");//reset des valeurs
#endif
            adresse_Numero = 0;
            digit_C = 0;
            digit_D = 0;
            digit_U = 0;
            digit_COMMAND = 0;
            changement_adresse = 0;
            menu = 1; //retour en mode RUN
            vit_abs = 0;
            lcd.clear();
            affichageMenu(1);
          }//  END IF appui sur DELETE
        }//end if bouton_valeur=20 ou changement_adresse=1
        if (bouton_valeur != 30)
        {
          affichage_LCD_INT(3, 1, comptageAdresse + 1);
          affichage_LCD_INT(7, 1, LocoAdresse[comptageAdresse]);
          affichage_LCD_INT(14, 1, digit_COMMAND);
        }

      }//end if comptageAdresse<maxLocos

    }// END IF menu = 3

    My_Receiver.resume();
    time_button = millis();
    //refresh des fonctions type bascule (ON/OFF)

#ifdef mode_debug
    Serial.println(" ");
    Serial.print(time);
    Serial.print("  ");
    Serial.println(millis());
#endif

  }// fin de controle reception commande IR

  //rafraichissement regulier des commandes de fonction
  if ((time + timerefresh) < millis())//instruction de rafraichissement regulier des commandes de fonction
  {
    //confirmer la valeur CommandeFonction de chaque loco
    for (int i = 0; i <= maxLocos; i++)
    {
#ifndef affichage_code //evite d'afficher les commandes a destination de la centrale lorsque l'on affiche les codes
      Serial.print("<f ");
      Serial.print(LocoAdresse[i]);
      Serial.print(" ");
      Serial.print(CommandeFonction[i]);
      Serial.write(" >");
      Serial.println(" ");//provoque l'affichage des commandes en colonne
#endif
    }//fin de FOR (int i = 0; i <= maxLocos; i++)

    time = millis();
  }// fin de rafraichissement if(time+timerefresh)

} //fin de LOOP

//fonction de determination de la vitesse de la loco
void LocoVitesse()
{
  switch (bouton_valeur)
  {
    case 0:
      vit_abs = 0;
      break;

    case 1:
      vit_abs = 15;
      break;

    case 2:
      vit_abs = 25;
      break;

    case 3:
      vit_abs = 35;
      break;

    case 4:
      vit_abs = 45;
      break;

    case 5:
      vit_abs = 55;
      break;

    case 6:
      vit_abs = 65;
      break;

    case 7:
      vit_abs = 75;
      break;

    case 8:
      vit_abs = 85;
      break;

    case 9:
      vit_abs = 95;
      break;

    case 10:// fleche gauche - sens de marche AR
      vit_abs = 0;//forcer l'arret avant le changement de sens de marche
      sens_marche = 0;
      affichage_sens_marche(0);
      bip(tone_pin, frequence_A1, duree_courte, 1);
      break;

    case 11:// fleche droite - sens de marche AV
      vit_abs = 0;// forcer l'arret avant de changer de sens de marche
      sens_marche = 1;
      affichage_sens_marche(1);
      bip(tone_pin, frequence_A1, duree_courte, 1);
      break;

    case 12://fleche haut
      //limitation de montee a 127
      if (vit_abs < 126) {
        vit_abs = vit_abs + 1;
      }
      break;

    case 13://fleche bas
      //limitation de descente à 0
      if (vit_abs >= 1) {
        vit_abs = vit_abs - 1;
      }
      break;
  }

  //commande de vitesse DCC++

  LocoDirection[AdresseActive] = sens_marche; //enregistrement du sens de marche dans la loco active
  LocoSpeed[AdresseActive] = vit_abs; //enregistrement de la vitesse dans la loco active

  //affichage des parametres envoyes lors de la commande de vitesse
#ifdef mode_debug
  Serial.print ("Adresse loco : ");
  Serial.print (LocoAdresse[AdresseActive]);
  Serial.print (" Num registre loco : ");
  Serial.print (AdresseActive + 1);
  Serial.print (" Vitesse : ");
  Serial.print (LocoSpeed[AdresseActive]);
  Serial.print (" Direction : ");
  if (LocoDirection[AdresseActive] == 1)
  {
    Serial.println("AV");
  }
  else
  {
    Serial.println("AR");
  }
#endif

   Serial.print("<t ");
  //on utilise comme registre le numero d'ordre de la loco active + 1 (AdresseActive = 0 => registre = 1; AdresseActive = 3 => registre = 4)
  Serial.print(AdresseActive + 1);
  Serial.print(" ");
  Serial.print(LocoAdresse[AdresseActive]);
  Serial.print(" ");
  Serial.print(LocoSpeed[AdresseActive]);
  Serial.print(" ");
  Serial.print(LocoDirection[AdresseActive]);
  Serial.write(">");
}

void affichageMenu(byte num_menu)
{
  lcd.clear();
  lcd.setCursor (8, 1);
  if (num_menu == 0)//bouton Stop ou Pause
  {
    lcd.clear();
    lcd.print("STOP");
    lcd.setCursor (0, 1);
    lcd.print(ver);
    vit_abs = 0;
    //remise a 0 des valeurs de vitesse des locos en cas de Stop ou Pause
    for (int i = 0; i < maxLocos; i++)
    {
      LocoSpeed [i] = 0;
      LocoVitesse();
      Serial.print (LocoSpeed[i]);
    }
    Serial.print("<0>");// coupure signal DCC
  }
  if (num_menu == 1)
  {
    Serial.print("<1>");
    lcd.setCursor (0, 0);
    lcd.print("L");
    lcd.setCursor (0, 1);
    lcd.print("R");
    affichage_LCD_INT(4, 0, LocoAdresse[AdresseActive]);
    lcd.setCursor(11, 0);
    lcd.print("F1234");
    AffichageFonction();

  }
  if (num_menu == 2)
  {
    lcd.setCursor (0, 0);
    lcd.print(">");
    lcd.setCursor(4, 0);
    lcd.print("?");
    lcd.setCursor (0, 1);
    lcd.print("P");
    lcd.setCursor(6, 0);
    lcd.print("CV NUM");
    lcd.setCursor(6, 1);
    lcd.print("CV VAL");
  }

  if (num_menu == 3)
  {
    lcd.setCursor (1, 0);
    lcd.print("LOC NOW CHANGE?");
    lcd.setCursor(0, 1);
    lcd.print("A");
  }
}

void affichage_LCD_INT(int colonne_LCD, int ligne_LCD, int valeur_affichage)
{
  /* fonction permettant l'affichage d'une valeur entière (INT) jusqu'à 999 sur le LCD
   colonne_LCD = position du chiffre des unités (min = 2, max=15)
   ligne_LCD = ligne d'ecriture (ligne superieure = 0, ligne inferieure =1
   valeur_affichage = de type INT
   */

  if (colonne_LCD<2 or colonne_LCD>15)
  {
    colonne_LCD = 2;
  }

  if (ligne_LCD<0 or ligne_LCD>1)
  {
    ligne_LCD = 0;
  }

  if (valeur_affichage < 10)
  {
    lcd.setCursor(colonne_LCD - 2, ligne_LCD);
    lcd.print ("  ");
    lcd.setCursor(colonne_LCD, ligne_LCD);
  }
  if (valeur_affichage > 9 and valeur_affichage < 100)
  {
    lcd.setCursor(colonne_LCD - 2, ligne_LCD);
    lcd.print (" ");
    lcd.setCursor(colonne_LCD - 1, ligne_LCD);
  }
  if (valeur_affichage >= 100)
  {
    lcd.setCursor(colonne_LCD - 2, ligne_LCD);
  }
  lcd.print (valeur_affichage);
}

void affichage_sens_marche(int s_marche)
//sens de marche : 0 = arriere; 1 = avant
{
  lcd.setCursor (6, 1);
  lcd.print("  ");//effacement sens precedent

  lcd.setCursor (8, 1);
  lcd.print("  ");
  if (s_marche == 0)
  {
    lcd.setCursor(6, 1);
    lcd.print("<");
  }
  else // if (s_marche == 1);
  {
    lcd.setCursor(8, 1);
    lcd.print(">");
  }
}

void changementLoco(int autre_loco)
{
  if (menu == 1)
  {
    lcd.clear();
    AdresseActive = autre_loco;
    sens_marche = LocoDirection[AdresseActive];
    vit_abs = LocoSpeed[AdresseActive];
    affichageMenu(1);
    bip(tone_pin, frequence_G1, duree_moyenne, autre_loco + 1);// nombre de bip = numero d'ordre de la loco
    menu = 1; //retour au menu RUN
  }
}

void bip(int pin, unsigned int frequence, unsigned long duree, int nombre_repetition) // fonction de production des bips sonores
{
  unsigned long t0 = millis();
  for (int i = 0; i < nombre_repetition; i++)
  {
    NewTone(pin, frequence, duree);
    unsigned long t2 = millis();
    while (millis() - t2 < 150) //boucle d'attente de 150msec entre deux repetition de bip
    {
    }// fin de while

  }//fin de for

}// fin de void bip()

void Activation_Fonction(int num_fonction)
{
  //appui sur un bouton de fonction avec effet bascule :0 = appui; 1 = pas appui

  if (!LocoFonction[AdresseActive][num_fonction])
  {
    //la fonction est OFF =>activation
    //commande lights ON
    LocoFonction[AdresseActive][num_fonction] = true; //memorisation activation
    AffichageFonction();
  }
  else if (LocoFonction[AdresseActive][num_fonction])
  {
    //la fonction est ON =>extinction
    //commande lights ON
    LocoFonction[AdresseActive][num_fonction] = false;	//memorisation desactivation
    AffichageFonction();
  }
  bouton_valeur = 99; //reset etat bouton

  //calcul de CommandeFonction
  CommandeFonction[AdresseActive] = 128 ;//nombre de base pour F0 à F4
  for (int i = 0; i <= maxFonctions; i++)//calcul pour toutes les fonctions du groupe
  {
    if (LocoFonction[AdresseActive][i]) //addition uniquement si la fonction est ON
    {
      switch (i)
      {
        case 0:
          CommandeFonction[AdresseActive] = CommandeFonction[AdresseActive] + 16;
          break;

        case 1:
          CommandeFonction[AdresseActive] = CommandeFonction[AdresseActive] + 1;
          break;

        case 2:
          CommandeFonction[AdresseActive] = CommandeFonction[AdresseActive] + 2;
          break;

        case 3:
          CommandeFonction[AdresseActive] = CommandeFonction[AdresseActive] + 4;
          break;

        case 4:
          CommandeFonction[AdresseActive] = CommandeFonction[AdresseActive] + 8;
          break;

        default:
          CommandeFonction[AdresseActive] = 128;
          break;

      }// END Switch i

    }//Fin de IF LocoFonction[AdresseActive][i]
  }//fin de FOR (int i = 0; i <= maxFonctions; i++)

  Serial.print("<f ");
  Serial.print(LocoAdresse[AdresseActive]);
  Serial.print(" ");
  Serial.print(CommandeFonction[AdresseActive]);//transformer en caracteres?
  Serial.write(" >");
}	//END VOID Activation_Fonction

void AffichageFonction()
{
  //affichage de l'état des fonction

  for (int i = 0; i <= maxFonctions; i++)//calcul pour toutes les fonctions du groupe
  {
    lcd.setCursor(11 + i, 1);
    if (!LocoFonction[AdresseActive][i])//la fonction est désactivée
    {
      lcd.print ("_");//affichage fonction OFF
    }
    else if (LocoFonction[AdresseActive][i])//la fonction est activée
    {
      lcd.print ("|");//affichage fonction ON
    }
  }//Fin de FOR (int i = 0; i <= maxFonctions; i++)

}	//END VOID AffichageFonction

void lectureAdresse()//lecture des adresses dans l'EEPROM
{
  for (int i = 0; i < maxLocos; i++)//lecture adresse de toutes les locos enregistrées
  {
    LocoAdresse[i] = EEPROM[i];
  }

}//END VOID lectureAdresse()

void ecritureAdresse(int numero_loco)//ecriture de l'adresse dans l'EEPROM
{
  EEPROM[numero_loco] = LocoAdresse[numero_loco];
}//END VOID ecritureAdresse()
